@@ -95,6 +95,9 @@ typedef struct evalinfo_T {
     char *savelocale;    /* original LC_NUMERIC locale */
 } evalinfo_T;
 
+static void evaluate(
+	const wchar_t *exp, value_T *result, evalinfo_T *info, bool coerce)
+    __attribute__((nonnull));
 static void parse_assignment(evalinfo_T *info, value_T *result)
     __attribute__((nonnull));
 static bool do_assignment(const word_T *word, const value_T *value)
@@ -153,25 +156,14 @@ static void next_token(evalinfo_T *info)
 /* Evaluates the specified string as an arithmetic expression.
  * The argument string is freed in this function.
  * The result is converted into a string and returned as a newly-malloced
- * string. On error, an error message is printed to stderr and NULL is
- * returned. */
+ * string. On error, an error message is printed to the standard error and NULL
+ * is returned. */
 wchar_t *evaluate_arithmetic(wchar_t *exp)
 {
     value_T result;
     evalinfo_T info;
 
-    info.exp = exp;
-    info.index = 0;
-    info.parseonly = false;
-    info.error = false;
-    info.savelocale = xstrdup(setlocale(LC_NUMERIC, NULL));
-
-    next_token(&info);
-    parse_assignment(&info, &result);
-    if (posixly_correct)
-	coerce_number(&info, &result);
-
-    free(info.savelocale);
+    evaluate(exp, &result, &info, posixly_correct);
 
     wchar_t *resultstr;
     if (info.error) {
@@ -197,17 +189,7 @@ bool evaluate_index(wchar_t *exp, ssize_t *valuep)
     value_T result;
     evalinfo_T info;
 
-    info.exp = exp;
-    info.index = 0;
-    info.parseonly = false;
-    info.error = false;
-    info.savelocale = xstrdup(setlocale(LC_NUMERIC, NULL));
-
-    next_token(&info);
-    parse_assignment(&info, &result);
-    coerce_number(&info, &result);
-
-    free(info.savelocale);
+    evaluate(exp, &result, &info, true);
 
     bool ok;
     if (info.error) {
@@ -239,6 +221,23 @@ bool evaluate_index(wchar_t *exp, ssize_t *valuep)
     return ok;
 }
 
+void evaluate(
+	const wchar_t *exp, value_T *result, evalinfo_T *info, bool coerce)
+{
+    info->exp = exp;
+    info->index = 0;
+    info->parseonly = false;
+    info->error = false;
+    info->savelocale = xstrdup(setlocale(LC_NUMERIC, NULL));
+
+    next_token(info);
+    parse_assignment(info, result);
+    if (coerce)
+	coerce_number(info, result);
+
+    free(info->savelocale);
+}
+
 /* Parses an assignment expression.
  *   AssignmentExp := ConditionalExp
  *                  | ConditionalExp AssignmentOperator AssignmentExp */
@@ -294,7 +293,7 @@ void parse_assignment(evalinfo_T *info, value_T *result)
 bool do_assignment(const word_T *word, const value_T *value)
 {
     wchar_t *vstr = value_to_string(value);
-    if (!vstr)
+    if (vstr == NULL)
 	return false;
 
     wchar_t name[word->length + 1];
@@ -320,11 +319,10 @@ wchar_t *value_to_string(const value_T *value)
 		wmemcpy(name, value->v_var.contents, value->v_var.length);
 		name[value->v_var.length] = L'\0';
 		const wchar_t *var = getvar(name);
-		return var ? xwcsdup(var) : NULL;
+		return (var != NULL) ? xwcsdup(var) : NULL;
 	    }
-	default:
-	    assert(false);
     }
+    assert(false);
 }
 
 /* Does unary or binary long calculation according to the specified operator
@@ -418,46 +416,46 @@ long do_double_comparison(tokentype_T ttype, double v1, double v2)
 void parse_conditional(evalinfo_T *info, value_T *result)
 {
     bool saveparseonly = info->parseonly;
-start:
-    parse_logical_or(info, result);
-    if (info->token.type == TT_QUESTION) {
-	bool cond, valid = true;
-	value_T dummy;
 
-	coerce_number(info, result);
+    for (;;) {
+	value_T dummy;
+	value_T *result2;
+
+	result2 = info->parseonly ? &dummy : result;
+	parse_logical_or(info, result2);
+	if (info->token.type != TT_QUESTION)
+	    break;
+
+	bool cond, valid = true;
+
+	coerce_number(info, result2);
 	next_token(info);
-	switch (result->type) {
-	    case VT_INVALID:  valid = false, cond = true;  break;
-	    case VT_LONG:     cond = result->v_long;       break;
-	    case VT_DOUBLE:   cond = result->v_double;     break;
+	switch (result2->type) {
+	    case VT_INVALID:  valid = false, cond = true;   break;
+	    case VT_LONG:     cond = result2->v_long;       break;
+	    case VT_DOUBLE:   cond = result2->v_double;     break;
 	    default:          assert(false);
 	}
-	if (cond) {
-	    parse_assignment(info, result);
-	    if (info->token.type == TT_COLON) {
-		next_token(info);
-		info->parseonly = true;
-		parse_conditional(info, &dummy);
-		info->parseonly = saveparseonly;
-	    } else {
-		xerror(0, Ngt("arithmetic: `%ls' is missing"), L":");
-		info->error = true;
-		result->type = VT_INVALID;
-	    }
-	} else {
-	    info->parseonly = true;
-	    parse_assignment(info, &dummy);
-	    info->parseonly = saveparseonly;
-	    if (info->token.type == TT_COLON) {
-		next_token(info);
-		goto start;
-	    } else {
-		xerror(0, Ngt("arithmetic: `%ls' is missing"), L":");
-		info->error = true;
-		result->type = VT_INVALID;
-	    }
+
+	bool saveparseonly2 = info->parseonly || !valid;
+	info->parseonly = saveparseonly2 || !cond;
+
+	result2 = info->parseonly ? &dummy : result;
+	parse_assignment(info, result2);
+	if (info->token.type != TT_COLON) {
+	    xerror(0, Ngt("arithmetic: `%ls' is missing"), L":");
+	    info->error = true;
+	    result->type = VT_INVALID;
+	    break;
 	}
+
+	next_token(info);
+	info->parseonly = saveparseonly2 || cond;
     }
+
+    info->parseonly = saveparseonly;
+    if (info->parseonly)
+	result->type = VT_INVALID;
 }
 
 /* Parses a logical OR expression.
@@ -465,27 +463,28 @@ start:
 void parse_logical_or(evalinfo_T *info, value_T *result)
 {
     bool saveparseonly = info->parseonly;
+
     parse_logical_and(info, result);
     while (info->token.type == TT_PIPEPIPE) {
-	bool lhs, INIT(value), valid = true;
+	bool value, valid = true;
+
 	coerce_number(info, result);
 	next_token(info);
 	switch (result->type) {
-	    case VT_INVALID: valid = false, lhs = true;  break;
-	    case VT_LONG:    lhs = result->v_long;       break;
-	    case VT_DOUBLE:  lhs = result->v_double;     break;
+	    case VT_INVALID: valid = false, value = true;  break;
+	    case VT_LONG:    value = result->v_long;       break;
+	    case VT_DOUBLE:  value = result->v_double;     break;
 	    default:         assert(false);
 	}
-	info->parseonly |= lhs;
+
+	info->parseonly |= value;
 	parse_logical_and(info, result);
 	coerce_number(info, result);
-	if (!lhs) switch (result->type) {
+	if (!value) switch (result->type) {
 	    case VT_INVALID: valid = false;             break;
 	    case VT_LONG:    value = result->v_long;    break;
 	    case VT_DOUBLE:  value = result->v_double;  break;
 	    default:         assert(false);
-	} else {
-	    value = true;
 	}
 	if (valid)
 	    result->type = VT_LONG, result->v_long = value;
@@ -500,27 +499,28 @@ void parse_logical_or(evalinfo_T *info, value_T *result)
 void parse_logical_and(evalinfo_T *info, value_T *result)
 {
     bool saveparseonly = info->parseonly;
+
     parse_inclusive_or(info, result);
     while (info->token.type == TT_AMPAMP) {
-	bool lhs, INIT(value), valid = true;
+	bool value, valid = true;
+
 	coerce_number(info, result);
 	next_token(info);
 	switch (result->type) {
-	    case VT_INVALID: valid = false, lhs = false;  break;
-	    case VT_LONG:    lhs = result->v_long;        break;
-	    case VT_DOUBLE:  lhs = result->v_double;      break;
+	    case VT_INVALID: valid = false, value = false;  break;
+	    case VT_LONG:    value = result->v_long;        break;
+	    case VT_DOUBLE:  value = result->v_double;      break;
 	    default:         assert(false);
 	}
-	info->parseonly |= !lhs;
+
+	info->parseonly |= !value;
 	parse_inclusive_or(info, result);
 	coerce_number(info, result);
-	if (lhs) switch (result->type) {
+	if (value) switch (result->type) {
 	    case VT_INVALID: valid = false;             break;
 	    case VT_LONG:    value = result->v_long;    break;
 	    case VT_DOUBLE:  value = result->v_double;  break;
 	    default:         assert(false);
-	} else {
-	    value = false;
 	}
 	if (valid)
 	    result->type = VT_LONG, result->v_long = value;
@@ -639,7 +639,8 @@ void parse_equality(evalinfo_T *info, value_T *result)
 		    case VT_INVALID:
 			result->type = VT_INVALID;
 			break;
-		    case VT_VAR:     assert(false);
+		    case VT_VAR:
+			assert(false);
 		}
 		break;
 	    default:
@@ -718,7 +719,7 @@ void parse_shift(evalinfo_T *info, value_T *result)
     }
 }
 
-/* Parses a additive expression.
+/* Parses an additive expression.
  *   AdditiveExp := MultiplicativeExp
  *                | AdditiveExp "+" MultiplicativeExp
  *                | AdditiveExp "-" MultiplicativeExp */
@@ -745,7 +746,8 @@ void parse_additive(evalinfo_T *info, value_T *result)
 		    case VT_INVALID:
 			result->type = VT_INVALID;
 			break;
-		    case VT_VAR:     assert(false);
+		    case VT_VAR:
+			assert(false);
 		}
 		break;
 	    default:
@@ -797,7 +799,8 @@ void parse_multiplicative(evalinfo_T *info, value_T *result)
 		    case VT_INVALID:
 			result->type = VT_INVALID;
 			break;
-		    case VT_VAR:     assert(false);
+		    case VT_VAR:
+			assert(false);
 		}
 		break;
 	    default:
@@ -894,8 +897,10 @@ void parse_postfix(evalinfo_T *info, value_T *result)
 		    coerce_number(info, result);
 		    value_T value = *result;
 		    do_increment_or_decrement(info->token.type, &value);
-		    if (!do_assignment(&saveword, &value))
-			info->error = true, result->type = VT_INVALID;
+		    if (!do_assignment(&saveword, &value)) {
+			info->error = true;
+			result->type = VT_INVALID;
+		    }
 		} else if (result->type != VT_INVALID) {
 		    xerror(0, Ngt("arithmetic: "
 				"operator `%ls' requires a variable"),
@@ -951,8 +956,6 @@ void parse_primary(evalinfo_T *info, value_T *result)
 	    break;
 	case TT_NUMBER:
 	    parse_as_number(info, result);
-	    if (result->type == VT_INVALID)
-		info->error = true;
 	    next_token(info);
 	    break;
 	case TT_IDENTIFIER:
@@ -974,12 +977,12 @@ void parse_primary(evalinfo_T *info, value_T *result)
 void parse_as_number(evalinfo_T *info, value_T *result)
 {
     word_T *word = &info->token.word;
-    wchar_t w[word->length + 1];
-    wcsncpy(w, word->contents, word->length);
-    w[word->length] = L'\0';
+    wchar_t wordstr[word->length + 1];
+    wcsncpy(wordstr, word->contents, word->length);
+    wordstr[word->length] = L'\0';
 
     long longresult;
-    if (xwcstol(w, 0, &longresult)) {
+    if (xwcstol(wordstr, 0, &longresult)) {
 	result->type = VT_LONG;
 	result->v_long = longresult;
 	return;
@@ -989,8 +992,8 @@ void parse_as_number(evalinfo_T *info, value_T *result)
 	wchar_t *end;
 	setlocale(LC_NUMERIC, "C");
 	errno = 0;
-	doubleresult = wcstod(w, &end);
-	bool ok = !errno && !*end;
+	doubleresult = wcstod(wordstr, &end);
+	bool ok = (errno == 0 && *end == L'\0');
 	setlocale(LC_NUMERIC, info->savelocale);
 	if (ok) {
 	    result->type = VT_DOUBLE;
@@ -998,7 +1001,8 @@ void parse_as_number(evalinfo_T *info, value_T *result)
 	    return;
 	}
     }
-    xerror(0, Ngt("arithmetic: `%ls' is not a valid number"), w);
+    xerror(0, Ngt("arithmetic: `%ls' is not a valid number"), wordstr);
+    info->error = true;
     result->type = VT_INVALID;
 }
 
@@ -1010,21 +1014,22 @@ void coerce_number(evalinfo_T *info, value_T *value)
     if (value->type != VT_VAR)
 	return;
 
-    const wchar_t *v;
+    const wchar_t *varvalue;
     {
-	wchar_t name[value->v_var.length + 1];
-	wmemcpy(name, value->v_var.contents, value->v_var.length);
-	name[value->v_var.length] = L'\0';
-	v = getvar(name);
+	word_T *name = &value->v_var;
+	wchar_t namestr[name->length + 1];
+	wmemcpy(namestr, name->contents, name->length);
+	namestr[name->length] = L'\0';
+	varvalue = getvar(namestr);
     }
-    if (!v || !v[0]) {
+    if (varvalue == NULL || varvalue[0] == L'\0') {
 	value->type = VT_LONG;
 	value->v_long = 0;
 	return;
     }
 
     long longresult;
-    if (xwcstol(v, 0, &longresult)) {
+    if (xwcstol(varvalue, 0, &longresult)) {
 	value->type = VT_LONG;
 	value->v_long = longresult;
 	return;
@@ -1033,14 +1038,14 @@ void coerce_number(evalinfo_T *info, value_T *value)
 	double doubleresult;
 	wchar_t *end;
 	errno = 0;
-	doubleresult = wcstod(v, &end);
-	if (!errno && !*end) {
+	doubleresult = wcstod(varvalue, &end);
+	if (errno == 0 && *end == L'\0') {
 	    value->type = VT_DOUBLE;
 	    value->v_double = doubleresult;
 	    return;
 	}
     }
-    xerror(0, Ngt("arithmetic: `%ls' is not a valid number"), v);
+    xerror(0, Ngt("arithmetic: `%ls' is not a valid number"), varvalue);
     info->error = true;
     value->type = VT_INVALID;
     return;
@@ -1058,9 +1063,9 @@ void coerce_integer(evalinfo_T *info, value_T *value)
 }
 
 /* Unifies the types of the numeric values.
- * If a given value is VT_VAR, it is `coerce_number'ed.
- * Then, if one of the specified values is of VT_LONG and the other is of
- * VT_DOUBLE, the VT_LONG value is converted into VT_DOUBLE.
+ * First, the given values are `coerce_number'ed.
+ * Then, if one of the values is of VT_LONG and the other is of VT_DOUBLE, the
+ * VT_LONG value is converted into VT_DOUBLE.
  * If the both values are of VT_LONG, VT_LONG is returned without conversion.
  * If either value is of VT_DOUBLE and the other is not VT_INVALID, VT_DOUBLE
  * is returned after conversion. If either value is VT_INVALID, the return
@@ -1076,9 +1081,9 @@ valuetype_T coerce_type(evalinfo_T *info, value_T *value1, value_T *value2)
     assert(value1->type == VT_LONG || value1->type == VT_DOUBLE);
     assert(value2->type == VT_LONG || value2->type == VT_DOUBLE);
 
-    value_T *value = (value1->type == VT_LONG) ? value1 : value2;
-    value->type = VT_DOUBLE;
-    value->v_double = (double) value->v_long;
+    value_T *value_to_coerce = (value1->type == VT_LONG) ? value1 : value2;
+    value_to_coerce->type = VT_DOUBLE;
+    value_to_coerce->v_double = (double) value_to_coerce->v_long;
     return VT_DOUBLE;
 }
 
