@@ -105,6 +105,9 @@ extern void endhostent(void);
 #endif
 
 
+static void select_candidate(void selector(int offset), int offset)
+    __attribute__((nonnull));
+static void select_candidate_by_offset(int offset);
 static void free_candidate(void *c)
     __attribute__((nonnull));
 static void free_context(le_context_T *ctxt);
@@ -117,8 +120,6 @@ static void print_compopt_info(const le_compopt_T *compopt)
     __attribute__((nonnull));
 
 static void execute_completion_function(void);
-static bool autoload_completion_function(void);
-static bool call_standard_completion_function(void);
 static void complete_command_default(void);
 
 static void simple_completion(le_candgentype_T type);
@@ -150,6 +151,7 @@ static void word_completion(size_t count, ...)
 static size_t get_common_prefix_length(void)
     __attribute__((pure));
 static void update_main_buffer(bool subst, bool finish);
+static void insert_to_main_buffer(wchar_t c);
 static bool need_subst(void);
 static void substitute_source_word_all(void);
 static void quote(xwcsbuf_T *restrict buf,
@@ -281,10 +283,13 @@ void lecr_longest_common_prefix(void)
     le_complete_cleanup();
 }
 
-/* Increases `le_selected_candidate_index' by `offset', selecting the `offset'th
- * next candidate. If there are no candidates, simply calls `le_complete' to
- * produce candidates. */
-void le_complete_select_candidate(int offset)
+/* Updates `le_selected_candidate_index' by calling `selector' with the argument
+ * `offset'. The `selector' function must update `le_selected_candidate_index'
+ * according to the given `offset'.
+ * The main buffer contents is also updated so that it contains the value of the
+ * new selected candidate.
+ * If there are no candidates, `le_complete' is called to produce candidates. */
+void select_candidate(void selector(int offset), int offset)
 {
     if (le_candidates.contents == NULL) {
 	le_complete(lecr_normal);
@@ -293,6 +298,20 @@ void le_complete_select_candidate(int offset)
 	return;
     }
 
+    selector(offset);
+    update_main_buffer(false, false);
+}
+
+/* Increases `le_selected_candidate_index' by `offset', selecting the `offset'th
+ * next candidate. If there are no candidates, simply calls `le_complete' to
+ * produce candidates. */
+void le_complete_select_candidate(int offset)
+{
+    select_candidate(select_candidate_by_offset, offset);
+}
+
+void select_candidate_by_offset(int offset)
+{
     assert(le_selected_candidate_index <= le_candidates.length);
     if (offset >= 0) {
 	offset %= le_candidates.length + 1;
@@ -306,8 +325,6 @@ void le_complete_select_candidate(int offset)
 	    le_selected_candidate_index += le_candidates.length - offset + 1;
     }
     assert(le_selected_candidate_index <= le_candidates.length);
-
-    update_main_buffer(false, false);
 }
 
 /* Selects the first candidate of the `offset'th next column.
@@ -315,15 +332,7 @@ void le_complete_select_candidate(int offset)
  */
 void le_complete_select_column(int offset)
 {
-    if (le_candidates.contents == NULL) {
-	le_complete(lecr_normal);
-	return;
-    } else if (le_candidates.length == 0) {
-	return;
-    }
-
-    le_selected_candidate_index = le_display_select_column(offset);
-    update_main_buffer(false, false);
+    select_candidate(le_display_select_column, offset);
 }
 
 /* Selects the first candidate of the `offset'th next page.
@@ -331,18 +340,10 @@ void le_complete_select_column(int offset)
  */
 void le_complete_select_page(int offset)
 {
-    if (le_candidates.contents == NULL) {
-	le_complete(lecr_normal);
-	return;
-    } else if (le_candidates.length == 0) {
-	return;
-    }
-
-    le_selected_candidate_index = le_display_select_page(offset);
-    update_main_buffer(false, false);
+    select_candidate(le_display_select_page, offset);
 }
 
-/* If `index' is not positive, performs completion and list candidates.
+/* If `index' is not positive, performs completion and lists candidates.
  * Otherwise, substitutes the source word with the `index'th candidate and
  * cleans up.
  * Returns true iff the source word was successfully substituted. */
@@ -371,7 +372,7 @@ bool le_complete_fix_candidate(int index)
     if (!subst) {
 	const le_candidate_T *cand =
 	    le_candidates.contents[le_selected_candidate_index];
-	subst = !matchwcsprefix(cand->origvalue, ctxt->src);
+	subst = (matchwcsprefix(cand->origvalue, ctxt->src) == NULL);
     }
     update_main_buffer(subst, true);
     le_complete_cleanup();
@@ -413,7 +414,7 @@ void free_context(le_context_T *ctxt)
     }
 }
 
-/* Sorts the candidates in the candidate list and remove duplicates. */
+/* Sorts the candidates in the candidate list and removes duplicates. */
 void sort_candidates(void)
 {
     qsort(le_candidates.contents,
@@ -521,6 +522,9 @@ void print_context_info(const le_context_T *ctxt)
  * enabled. */
 void print_compopt_info(const le_compopt_T *compopt)
 {
+    if (!le_state_is_compdebug)
+	return;
+
     const char *INIT(s);
     le_compdebug("target word without prefix: \"%ls\"", compopt->src);
     for (const le_comppattern_T *p = compopt->patterns; p != NULL; p = p->next){
@@ -542,8 +546,9 @@ void print_compopt_info(const le_compopt_T *compopt)
 /* name of the file that is autoloaded in the first completion */
 #define INIT_COMPFILE "completion/INIT"
 
-/* default completion function name */
-#define DEFAULT_COMPFUNC "completion//default"
+/* completion function names */
+#define COMMAND_COMPFUNC  "completion//command"
+#define ARGUMENT_COMPFUNC "completion//argument"
 
 /* Loads and executes completion function to generate candidates. */
 void execute_completion_function(void)
@@ -561,16 +566,12 @@ void execute_completion_function(void)
 	    simple_completion(CGT_FILE);
 	    break;
 	case CTXT_COMMAND:
-	    if (!call_completion_function(L"" DEFAULT_COMPFUNC))
+	    if (!call_completion_function(L"" COMMAND_COMPFUNC))
 		complete_command_default();
 	    break;
 	case CTXT_ARGUMENT:
-	    if (!call_standard_completion_function()) {
-		if (autoload_completion_function())
-		    call_standard_completion_function();
-		else if (!call_completion_function(L"" DEFAULT_COMPFUNC))
-		    simple_completion(CGT_FILE);
-	    }
+	    if (!call_completion_function(L"" ARGUMENT_COMPFUNC))
+		simple_completion(CGT_FILE);
 	    break;
 	case CTXT_TILDE:
 	    simple_completion(CGT_LOGNAME);
@@ -599,70 +600,6 @@ void execute_completion_function(void)
 
 }
 
-/* Define the completion function by autoloading "completion/<ctxt->pwords[0]>".
- * Returns true if a file is autoloaded. */
-bool autoload_completion_function(void)
-{
-    const wchar_t *cmdname = ctxt->pwords[0];
-    if (cmdname == NULL)
-	/* We're completing the command name so there's nothing to autoload */
-	return false;
-
-    xwcsbuf_T filename;
-    wb_init(&filename);
-    wb_cat(&filename, L"completion/");
-    wb_cat(&filename, cmdname);
-
-    bool ok = autoload_completion_function_file(filename.contents, cmdname);
-    if (!ok) {
-	/* try the filename without the directory components */
-	const wchar_t *newcmdname = wcsrchr(cmdname, L'/');
-	if (newcmdname != NULL && *++newcmdname != L'\0') {
-	    wb_clear(&filename);
-	    wb_cat(&filename, L"completion/");
-	    wb_cat(&filename, newcmdname);
-	    ok = autoload_completion_function_file(filename.contents, cmdname);
-	}
-    }
-
-    wb_destroy(&filename);
-
-    return ok;
-}
-
-/* Calls standard completion function to generate candidates.
- * If a function whose name is "completion/<ctxt->pwords[0]>" has been defined,
- * that function is called with no arguments.
- * Returns false if the function has not been defined. */
-bool call_standard_completion_function(void)
-{
-    const wchar_t *cmdname = ctxt->pwords[0];
-    if (cmdname == NULL)
-	/* We're completing the command name so there's no standard function */
-	return false;
-
-    xwcsbuf_T funcname;
-    wb_init(&funcname);
-    wb_cat(&funcname, L"completion/");
-    wb_cat(&funcname, cmdname);
-
-    bool ok = call_completion_function(funcname.contents);
-    if (!ok) {
-	/* try the function name without the directory components */
-	cmdname = wcsrchr(cmdname, L'/');
-	if (cmdname != NULL && *++cmdname != L'\0') {
-	    wb_clear(&funcname);
-	    wb_cat(&funcname, L"completion/");
-	    wb_cat(&funcname, cmdname);
-	    ok = call_completion_function(funcname.contents);
-	}
-    }
-
-    wb_destroy(&funcname);
-
-    return ok;
-}
-
 /* Sets special local variables $WORDS and $TARGETWORD in the current variable
  * environment. Also sets the $IFS variable to the default value. */
 void set_completion_variables(void)
@@ -699,13 +636,13 @@ void complete_command_default(void)
 
     compopt.suffix = NULL;
     compopt.terminate = true;
-    if (wcschr(ctxt->src, L'/')) {
+    if (wcschr(ctxt->src, L'/') != NULL) {
 	// pattern1.next = NULL;
 	compopt.type = CGT_EXECUTABLE;
     } else {
 	pattern1.next = &pattern2;
 	compopt.type = CGT_COMMAND;
-	if (ctxt->quote == QUOTE_NORMAL && !wcschr(ctxt->pattern, L'\\'))
+	if (ctxt->quote == QUOTE_NORMAL && wcschr(ctxt->pattern, L'\\') == NULL)
 	    compopt.type |= CGT_KEYWORD | CGT_NALIAS;
     }
     print_compopt_info(&compopt);
@@ -764,8 +701,8 @@ void generate_candidates(const le_compopt_T *compopt)
  * The ignored prefix in `ctxt->origsrc' is prepended to the candidate value.
  * A description for the candidate can be given as `desc', which may be NULL
  * when no description is provided.
- * Arguments `value' and `desc' must be a freeable string, which is used as the
- * candidate value/description.
+ * Arguments `value' and `desc' must be freeable strings, which are used as the
+ * candidate value and description, respectively.
  * This function must NOT be used for a CT_FILE candidate.
  * If `value' is NULL, this function does nothing (except freeing `desc'). */
 void le_new_candidate(le_candtype_T type, wchar_t *value, wchar_t *desc,
@@ -820,24 +757,24 @@ void le_add_candidate(le_candidate_T *cand, const le_compopt_T *compopt)
     }
 
     cand->origvalue = wb_towcs(&buf);
-    cand->value = cand->origvalue + prefixlength;
+    cand->value = &cand->origvalue[prefixlength];
     cand->terminate = compopt->terminate && allowterminate;
 
     if (le_state_is_compdebug) {
 	const char *typestr = NULL;
 	switch (cand->type) {
-	    case CT_WORD:      typestr = "word";                       break;
-	    case CT_FILE:      typestr = "file";                       break;
-	    case CT_COMMAND:   typestr = "command";                    break;
-	    case CT_ALIAS:     typestr = "alias";                      break;
-	    case CT_OPTION:    typestr = "option";                     break;
-	    case CT_VAR:       typestr = "variable";                   break;
-	    case CT_JOB:       typestr = "job";                        break;
-	    case CT_SIG:       typestr = "signal";                     break;
-	    case CT_LOGNAME:   typestr = "user name";                  break;
-	    case CT_GRP:       typestr = "group name";                 break;
-	    case CT_HOSTNAME:  typestr = "host name";                  break;
-	    case CT_BINDKEY:   typestr = "lineedit command";           break;
+	    case CT_WORD:      typestr = "word";              break;
+	    case CT_FILE:      typestr = "file";              break;
+	    case CT_COMMAND:   typestr = "command";           break;
+	    case CT_ALIAS:     typestr = "alias";             break;
+	    case CT_OPTION:    typestr = "option";            break;
+	    case CT_VAR:       typestr = "variable";          break;
+	    case CT_JOB:       typestr = "job";               break;
+	    case CT_SIG:       typestr = "signal";            break;
+	    case CT_LOGNAME:   typestr = "user name";         break;
+	    case CT_GRP:       typestr = "group name";        break;
+	    case CT_HOSTNAME:  typestr = "host name";         break;
+	    case CT_BINDKEY:   typestr = "lineedit command";  break;
 	}
 	le_compdebug("new %s candidate \"%ls\"", typestr, cand->origvalue);
 	if (cand->desc != NULL)
@@ -1151,8 +1088,8 @@ void generate_candidates_from_words(
     if (!le_compile_cpatterns(compopt))
 	return;
 
-    for (; *words != NULL; words++) {
-	wchar_t *word = *words;
+    const wchar_t *word;
+    for (size_t i = 0; (word = words[i]) != NULL; i++) {
 	if (le_wmatch_comppatterns(compopt, word))
 	    le_new_candidate(type, xwcsdup(word),
 		    (description == NULL) ? NULL : xwcsdup(description),
@@ -1220,8 +1157,7 @@ size_t get_common_prefix_length(void)
 
     if (le_state_is_compdebug) {
 	wchar_t value[common_prefix_length + 1];
-	cand = le_candidates.contents[0];
-	wcsncpy(value, cand->origvalue, common_prefix_length);
+	wmemcpy(value, cand->origvalue, common_prefix_length);
 	value[common_prefix_length] = L'\0';
 	le_compdebug("candidate common prefix: \"%ls\"", value);
     }
@@ -1294,51 +1230,53 @@ void update_main_buffer(bool subst, bool finish)
 	case QUOTE_NORMAL:
 	    break;
 	case QUOTE_SINGLE:
-	    wb_ninsert_force(&le_main_buffer, le_main_index, L"'", 1);
-	    le_main_index += 1;
+	    insert_to_main_buffer(L'\'');
 	    break;
 	case QUOTE_DOUBLE:
-	    wb_ninsert_force(&le_main_buffer, le_main_index, L"\"", 1);
-	    le_main_index += 1;
+	    insert_to_main_buffer(L'"');
 	    break;
     }
 
-    if (finish) {
-	if (ctxt->type & CTXT_VBRACED) {
-	    wb_ninsert_force(&le_main_buffer, le_main_index, L"}", 1);
-	    le_main_index += 1;
-	} else if (ctxt->type & CTXT_EBRACED) {
-	    wb_ninsert_force(&le_main_buffer, le_main_index, L",", 1);
-	    le_main_index += 1;
-	}
-	if (ctxt->type & CTXT_QUOTED) {
-	    wb_ninsert_force(&le_main_buffer, le_main_index, L"\"", 1);
-	    le_main_index += 1;
-	}
-	switch (ctxt->type & CTXT_MASK) {
-	    case CTXT_NORMAL:
-	    case CTXT_COMMAND:
-	    case CTXT_ARGUMENT:
-	    case CTXT_VAR:
-	    case CTXT_ARITH:
-	    case CTXT_ASSIGN:
-	    case CTXT_REDIR:
-	    case CTXT_REDIR_FD:
-	    case CTXT_FOR_IN:
-	    case CTXT_FOR_DO:
-	    case CTXT_CASE_IN:
-	    case CTXT_FUNCTION:
-		if (ctxt->type & (CTXT_EBRACED | CTXT_VBRACED))
-		    break;
-		wb_ninsert_force(&le_main_buffer, le_main_index, L" ", 1);
-		le_main_index += 1;
-		break;
-	    case CTXT_TILDE:
-		wb_ninsert_force(&le_main_buffer, le_main_index, L"/", 1);
-		le_main_index += 1;
-		break;
-	}
+    if (!finish)
+	return;
+
+    if (ctxt->type & CTXT_VBRACED) {
+	insert_to_main_buffer(L'}');
+	return;
+    } else if (ctxt->type & CTXT_EBRACED) {
+	insert_to_main_buffer(L',');
+	return;
     }
+    if (ctxt->type & CTXT_QUOTED) {
+	insert_to_main_buffer(L'"');
+    }
+    switch (ctxt->type & CTXT_MASK) {
+	case CTXT_NORMAL:
+	case CTXT_COMMAND:
+	case CTXT_ARGUMENT:
+	case CTXT_VAR:
+	case CTXT_ARITH:
+	case CTXT_ASSIGN:
+	case CTXT_REDIR:
+	case CTXT_REDIR_FD:
+	case CTXT_FOR_IN:
+	case CTXT_FOR_DO:
+	case CTXT_CASE_IN:
+	case CTXT_FUNCTION:
+	    insert_to_main_buffer(L' ');
+	    break;
+	case CTXT_TILDE:
+	    insert_to_main_buffer(L'/');
+	    break;
+    }
+}
+
+/* Inserts the specified character to the main buffer (`le_main_buffer') at the
+ * current index (`le_main_index'). The index is increased by 1. */
+void insert_to_main_buffer(wchar_t c)
+{
+    wb_ninsert_force(&le_main_buffer, le_main_index, &c, 1);
+    le_main_index += 1;
 }
 
 /* Determines whether the source word should be substituted even if
@@ -1349,7 +1287,7 @@ bool need_subst(void)
 {
     for (size_t i = 0; i < le_candidates.length; i++) {
 	const le_candidate_T *cand = le_candidates.contents[i];
-	if (!matchwcsprefix(cand->origvalue, ctxt->src))
+	if (matchwcsprefix(cand->origvalue, ctxt->src) == NULL)
 	    return true;
     }
     return false;
@@ -1371,17 +1309,16 @@ void substitute_source_word_all(void)
     for (size_t i = 0; i < le_candidates.length; i++) {
 	const le_candidate_T* cand = le_candidates.contents[i];
 
-	quote(wb_clear(&buf), cand->origvalue, QUOTE_NORMAL);
+	quote(&buf, cand->origvalue, QUOTE_NORMAL);
 	wb_wccat(&buf, L' ');
-	wb_ninsert_force(&le_main_buffer, le_main_index,
-		buf.contents, buf.length);
-	le_main_index += buf.length;
     }
+    wb_ninsert_force(&le_main_buffer, le_main_index, buf.contents, buf.length);
+    le_main_index += buf.length;
     wb_destroy(&buf);
 }
 
 /* Quotes characters in the specified string that are not treated literally
- * according to `quotetype'.
+ * in quotation mode `quotetype'.
  * The result is appended to the specified buffer, which must have been
  * initialized by the caller. */
 void quote(xwcsbuf_T *restrict buf,
@@ -1395,32 +1332,29 @@ void quote(xwcsbuf_T *restrict buf,
 	    wb_cat(buf, s);
 	    return;
 	case QUOTE_NORMAL:
-	    while (*s != L'\0') {
-		if (*s == L'\n') {
+	    for (size_t i = 0; s[i] != L'\0'; i++) {
+		if (s[i] == L'\n') {
 		    wb_ncat_force(buf, L"'\n'", 3);
 		} else {
-		    if (wcschr(quotechars, *s) || iswspace(*s))
+		    if (wcschr(quotechars, s[i]) != NULL || iswspace(s[i]))
 			wb_wccat(buf, L'\\');
-		    wb_wccat(buf, *s);
+		    wb_wccat(buf, s[i]);
 		}
-		s++;
 	    }
 	    return;
 	case QUOTE_SINGLE:
-	    while (*s != L'\0') {
-		if (*s != L'\'')
-		    wb_wccat(buf, *s);
+	    for (size_t i = 0; s[i] != L'\0'; i++) {
+		if (s[i] != L'\'')
+		    wb_wccat(buf, s[i]);
 		else
 		    wb_ncat_force(buf, L"'\\''", 4);
-		s++;
 	    }
 	    return;
 	case QUOTE_DOUBLE:
-	    while (*s != L'\0') {
-		if (wcschr(CHARS_ESCAPABLE, *s))
+	    for (size_t i = 0; s[i] != L'\0'; i++) {
+		if (wcschr(CHARS_ESCAPABLE, s[i]) != NULL)
 		    wb_wccat(buf, L'\\');
-		wb_wccat(buf, *s);
-		s++;
+		wb_wccat(buf, s[i]);
 	    }
 	    return;
     }
@@ -1595,7 +1529,7 @@ dupopterror:
 	goto finish;
     }
 
-    void *const *words = argv + xoptind;
+    void *const *words = &argv[xoptind];
 
     /* treat `prefix' */
     const wchar_t *src = ctxt->src, *pattern = ctxt->pattern;
