@@ -682,7 +682,6 @@ int handle_traps(void)
 
     int signum = 0;
     bool save_sigint_received = sigint_received;
-    struct parsestate_T *parsestate = NULL;
     struct execstate_T *execstate = NULL;
     savelaststatus = laststatus;
 
@@ -701,10 +700,8 @@ int handle_traps(void)
 #if YASH_ENABLE_LINEEDIT
 		    le_suspend_readline();
 #endif
-		    if (parsestate == NULL) {
-			parsestate = save_parse_state();
+		    if (execstate == NULL)
 			execstate = save_execstate();
-		    }
 		    signum = handled_signal = s->no;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
@@ -725,10 +722,8 @@ int handle_traps(void)
 #if YASH_ENABLE_LINEEDIT
 		    le_suspend_readline();
 #endif
-		    if (parsestate == NULL) {
-			parsestate = save_parse_state();
+		    if (execstate == NULL)
 			execstate = save_execstate();
-		    }
 		    signum = handled_signal = sigrtmin + i;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
@@ -743,10 +738,8 @@ int handle_traps(void)
     sigint_received |= save_sigint_received;
     savelaststatus = -1;
     handled_signal = -1;
-    if (parsestate != NULL) {
-	restore_parse_state(parsestate);
+    if (execstate != NULL)
 	restore_execstate(execstate);
-    }
 #if YASH_ENABLE_LINEEDIT
     if (shopt_notifyle && (le_state & LE_STATE_SUSPENDED))
 	print_job_status_all();
@@ -1205,9 +1198,7 @@ bool print_trap(const wchar_t *signame, const wchar_t *command)
 	return true;
 
     wchar_t *q = quote_sq(command);
-    bool ok = printf("trap -- %ls %ls\n", q, signame) >= 0;
-    if (!ok)
-	xerror(errno, Ngt("cannot print to the standard output"));
+    bool ok = xprintf("trap -- %ls %ls\n", q, signame);
     free(q);
     return ok;
 }
@@ -1393,24 +1384,16 @@ print_usage:
  * Returns false iff an IO error occurred. */
 bool print_signal(int signum, const wchar_t *name, bool verbose)
 {
-    int r;
-
     if (!verbose) {
-	r = printf("%ls\n", name);
+	return xprintf("%ls\n", name);
     } else {
 #if HAVE_STRSIGNAL
 	const char *sigdesc = strsignal(signum);
-	if (sigdesc)
-	    r = printf("%d\t%-10ls %s\n", signum, name, sigdesc);
+	if (sigdesc != NULL)
+	    return xprintf("%d\t%-10ls %s\n", signum, name, sigdesc);
 	else
 #endif
-	    r = printf("%d\t%-10ls\n", signum, name);
-    }
-    if (r >= 0) {
-	return true;
-    } else {
-	xerror(errno, Ngt("cannot print to the standard output"));
-	return false;
+	    return xprintf("%d\t%-10ls\n", signum, name);
     }
 }
 
